/*
*/
(function(){
 'use strict';
 const Ld = sLoader;

 Ld.module('exifReader',function(){

  let SObject = Ld.use('sobject');
  let M       = Ld.ref('main');
  let U       = Ld.use('utility');

  const makerList = [
   'olympus',
  ];

  const ExifReader = SObject.derive();
  this.body = ExifReader;

  function trim(s){ return s.trim(); }

  function rationalToFloat(r){
   // r[1]>0
   return r[0]/r[1];
  }

  // only ASCII
  ExifReader.dataViewToString = function(data,maxlen){
   let n = data.byteLength;
   if ( maxlen && n > maxlen ) { n = maxlen; }
   let s = '';
   for( let i = 0; i < n; i++ ) {
    s += String.fromCharCode(data.getUint8(i));
   }
   return s;
  };


  const exifTagTable = {
   IFD0: {
    0x0100: { name:'ImageWidth' },      // u32 -> jpegの場合exifには入ってない
    0x0101: { name:'ImageHeight' },     // u32 ->  
    0x010f: { name:'Make', conv:trim },            // string
    0x0110: { name:'Model', conv:trim },           // string
    0x0112: { name:'Orientation' },     // u16 1:Horizontal...
    0x9003: { name:'DateTimeOriginal' }, // string
    0x8769: { type:'IFD', group:'IFD0' },  // ExifIFD

    // ExifIFD
    0x829a: { name:'ExposureTime', conv: rationalToFloat },
    0x829d: { name:'FNumber', conv:rationalToFloat },
    0x8822: { name:'ExposureProgram' },
    0x8827: { name:'ISO' },
    0x8833: { name:'ISOSpeed' },
    0x8825: { type:'IFD', group:'GPSTags' },
    0x9201: { name:'ShutterSpeedValue', 
              conv: val=>val[1]>0? Math.pow(2,-val[0]/val[1]): 0
            },
    0x9202: { name:'ApertureValue' },
    0x9204: { name:'ExposureCompensation' },
    0x9209: { name:'Flash' },
    0x920a: { name:'FocalLength', conv:rationalToFloat },
    0x927c: { type:'MakerNote' },
    0xa402: { name:'ExposureMode' },
    0xa403: { name:'WhiteBalance' },
    0xa404: { name:'DigitalZoomRatio', conv:rationalToFloat }, // ur64
    0xa405: { name:'FocalLengthIn35mmFormat' },
    0xa433: { name:'LensMake' },
    0xa434: { name:'LensModel' },
    0xa500: { name:'Gamma' },
   },

   GPSTags:{
    0x0000: { name:'GPSVersionID', conv: d=>d.join('.') },
    0x0001: { name:'GPSLatitudeRef' },
    0x0002: { name:'_GPSLatitude' },
    0x0003: { name:'GPSLongitudeRef' },
    0x0004: { name:'_GPSLongitude' },
    0x0005: { name:'GPSAltitudeRef' },
    0x0006: { name:'_GPSAltitude', conv:rationalToFloat },
   }
  };


  ExifReader.processExif = function( exif,
    data, offset,
    table, group, littleEndian ) {

   let buf = data;
   let ifdque = [];
   let tagTable = table;

   function get_u8(p)  { return buf.getUint8(p); }

// 調査
   function get_c(p)   { return String.fromCharCode(buf.getUint8(p)); }
                             // only supported ASCII code.

   function get_u16(p) { return buf.getUint16(p,littleEndian); }
   function get_u32(p) { return buf.getUint32(p,littleEndian); }
   function get_s8(p)  { return buf.getInt8(p); }
   function get_s16(p) { return buf.getInt16(p,littleEndian); }
   function get_s32(p) { return buf.getInt32(p,littleEndian); }
   function get_ur64(p){ return [ get_u32(p), get_u32(p+4) ]; }
   function get_sr64(p){ return [ get_s32(p), get_u32(p+4) ]; }

   const dataSize = {
    1:  1,
    2:  1,
    3:  2,
    4:  4,
    5:  8,
    6:  1,
    7:  1,
    8:  2,
    9:  4,
   10: 16,
   /* 11:4, 12:8 */
   13:  4,  // IFD pointer?
   };

   const readProc = {
    1: get_u8,
    3: get_u16,
    4: get_u32,
    5: get_ur64,
    6: get_s8,
    8: get_s16,
    9: get_s32,
    10: get_sr64,
    /* unsupported 11:f32, 12:f64 */
    13: get_u32,
   };

   function readValue(p){
    let format = get_u16(p+2);
    let count = get_u32(p+4);
    let elsize = dataSize[format];
    let value, q;
    if ( count*elsize <= 4 ) {
     q = p+8;
    } else {
     q = get_u32(p+8);
    }
    if ( format === 2 ) { // string
     let s = '';
     for( let i = 0; i < count-1; i++ ) {
      s += get_c(q+i);
     }
     value = s;
    } else if ( format === 7 ) { // undefined
     value = new DataView(buf.buffer,buf.byteOffset+q,count);

    } else {
     let proc = readProc[format];
     if ( !proc ) {
      M.error('unsupported ',format);
     }
     else {
      let values = [];
      for( let i = 0; i < count; i++, q += elsize ) {
       values.push(proc(q));
      }
      if ( count === 1 ) {
       value = values[0];
      } else {
       value = values;
      }
     }
    }
    return value;
   }

   function makerNote(data){
    for( let i = 0; i < makerList.length; i++ ) {
     let plugin = Ld.use('exif-'+makerList[i]);
     if ( plugin.readMakerNote(exif,data) ) {
      break;
     }
    }
   }

   function readIfd(addr,group){
    let p = addr;
    let tbl = tagTable[group];
    while( p !== 0 ) {
     let ntag = get_u16(p);
     p += 2;
     for( let j = 0; j < ntag; j++, p += 12 ) {
      let tagId = get_u16(p);
      let tagInfo = tbl[tagId];
      if ( tagInfo === undefined ) { continue; }
      let tagName = tagInfo.name;
      let value = readValue(p);
      if ( tagInfo.conv ) {
       value = tagInfo.conv(value);
      }
      switch(tagInfo.type){
      case 'IFD':
       ifdque.push({ addr: value, group: tagInfo.group });
       break;
      case 'MakerNote': makerNote(value); break;
      default: exif[tagName] = value; break;
      }
     }
     p = get_u32(p);
    }
   }

   // processExif

   ifdque.push({ addr:offset, group:group });
   let ifd;
   while( (ifd = ifdque.shift() ) !== undefined ) {
    readIfd( ifd.addr, ifd.group );
   }

   return exif;

  };


  function makeSupplementalTags(exif){

   // GPSPosition
   if ( exif._GPSLatitude && exif.GPSLatitudeRef
    && exif._GPSLongitude && exif.GPSLongitudeRef ) {
    let d = exif._GPSLatitude;
    let lat = d[0][0]/d[0][1] + d[1][0]/d[1][1]/60 + d[2][0]/d[2][1]/3600;
    if ( exif.GPSLatitudeRef.match(/^s/i) ) {
// +-?
     lat = -lat;
    }
    exif.GPSLatitude = lat;
    d = exif._GPSLongitude;
    let lon = d[0][0]/d[0][1] + d[1][0]/d[1][1]/60 + d[2][0]/d[2][1]/3600;
    if ( exif.GPSLongitudeRef.match(/^w/i) ) {
     lon = -lon;
    }
    exif.GPSLongitude = lon;
   }

   // GPSAltitude
   if ( !isNaN(exif._GPSAltitude) && !isNaN(exif.GPSAltitudeRef) ) {
    exif.GPSAltitude = (exif.GPSAltitudeRef===1? -1: 1)*exif._GPSAltitude;
   }

  }


  //
  const readBuf = (file,pos,len)=>{
   if ( pos+len > file.size ) {
    throw('no exif data');
   }
   return new Promise((resolve, reject) => {
    const reader = new FileReader();
    reader.onerror = () => {
     reader.abort();
     reject(reader.error);
    };
    reader.onload = () => {
     resolve(new DataView(reader.result));
    };
    reader.readAsArrayBuffer(file.slice(pos,pos+len));
   });
  };



  // 
  ExifReader.readP = async(file)=>{
   let exif = {};

   try{
    let data;
    let pos = 0;
    data = await readBuf(file,pos,2);
    pos += 2;
    if ( data.getUint16(0) !== 0xFFD8 ) {
     throw('not jpeg');
    }

    // imageWidth, imageHeight
    for(;data = await readBuf(file,pos,9);){
     if ( data.getUint8(0) !== 0xFF ) {
      throw 'wrong#1';
     } else {
      let c = data.getUint8(1);
      if ( c >= 0xC0 && c <= 0xCF &&
          (c !== 0xC4 && c !== 0xC8 && c !== 0xCC) ) {
       exif.ImageWidth = data.getUint16(5);
       exif.ImageHeight = data.getUint16(7);
       break;
      } else {
       pos += 2+data.getUint16(2);
      }
     }
    }

    // seek App1
    pos = 2;
    for(;data = await readBuf(file,pos,4);){
     if ( data.getUint8(0) !== 0xFF ) {
      throw 'wrong#2';
     } else {
      if ( data.getUint8(1) === 0xE1 ) {
       // APP1 Marker 
       let len = data.getUint16(2);
       data = await readBuf(file,pos,len+2);
       break;
      }
      else{
       pos += 4+data.getUint16(2);
      }
     }
    }

    // App1


    // Exif Header 'Exif'
    if ( data.getUint32(4) !== 0x45786966 || data.getUint16(8) !== 0 ) {
     throw('wrong#3');
    }

    // TIFF 
 // byteOffset?
    data = new DataView(data.buffer,data.byteOffset+10,data.byteLength-10);

    let littleEndian;
    let byteOrder = data.getUint16(0,false);
    if ( byteOrder === 0x4D4D ) { // 'MM'
     littleEndian = false;
    } else if ( byteOrder === 0x4949 ) { // 'II'
     littleEndian = true;
    } else {
     throw('wrong#4');
    }
    ExifReader.processExif(
     exif, data, data.getUint32(4,littleEndian),
     exifTagTable, 'IFD0', littleEndian
    );
    makeSupplementalTags(exif);
   }
   catch(e){
    exif.error = e;
   }
   return exif;
  };


 });

})();
