//
//

(function(){
 'use strict';
 const Ld = sLoader;
 Ld.module('fov',function(){

  const Fov = {};

  Fov.compute = compute;
  Fov.getList = getList;
  Fov.getKey = getKey;

  Fov.getProfiles = getProfiles;
  Fov.restoreProfiles = restoreProfiles;
  Fov.reset = reset;
  Fov.clear = clear;
  Fov.addProfile = addProfile;


  const M = Ld.ref('main');
  const ExifReader = Ld.ref('exifReader');
  const BP = Ld.use('builtinProfiles');

  
  let files = {};
  let profiles = {};


  //  linear interpolation
  function interpolate(x0,y0,x1,y1,x){
   return (y0*(x1-x)+y1*(x-x0))/(x1-x0);
  }

// zero div?
  // focus
  function interpolateF(fsc0,fov0,fsc1,fov1,fsc){
   return 1/interpolate(fsc0,1/fov0,fsc1,1/fov1,fsc);
  }

// zero div?
  // zoom(linear)
  function interpolateZ(zsc0,fov0,zsc1,fov1,zsc){
   return 1/interpolate(zsc0,1/fov0,zsc1,1/fov1,zsc);
  }

  // zoom(log)
  function interpolateZLog(zsc0,fov0,zsc1,fov1,zsc){
   return 1/Math.exp(
    interpolate(zsc0,Math.log(1/fov0),zsc1,Math.log(1/fov1),zsc));
  }


  function convertData(profile,fields){
   // try!!
   if ( profile.fields === undefined ) {
    // 配列でない場合とかも!!
    M.error('invalid profile data');
    return;
   }
   let idx0 = {};
   for( let i = 0; i < profile.fields.length; i++ ) {
    idx0[profile.fields[i]] = i;
   }
   let idx = [];
   for( let i = 0; i < fields.length; i++ ) {
    idx[i] = idx0[fields[i]];
    if ( idx[i] === undefined ) {
     // メッセージ
//console.log(profile);
     M.error('invalid profile data');
     return;
    }
   }
   let data = [];
   for( let i = 0; i < profile.data.length; i++ ) {
    let r = {};
    for( let j = 0; j < fields.length; j++ ) {
     r[fields[j]] = profile.data[i][idx[j]];
    }
    data[i] = r;
   }
   return data;
  }


  function fixedFocal(profile,exif){
   let fsc = exif.FocusStepCount;

// 1
   if ( exif.DigitalZoomRatio == 0 ) {
    M.log('DigitalZoomRatio');
    return;
   }
   let data = convertData(profile,['fsc','fov']);
   if ( !data || data.length < 2 ) return;
   data = data.sort((a,b)=>a.fsc-b.fsc);
   let i = 0;
   for( ; i < data.length-1 && data[i+1].fsc < fsc ; i++ ) {}
   if ( i >= data.length-1 ) i--;
   let r0 = data[i], r1 = data[i+1];
   // r0.fsc != f1.fsc
   let fov = interpolateF(r0.fsc,r0.fov,r1.fsc,r1.fov,fsc);
   return fov;
  }


  function zoomLens(profile,exif){
   let zsc  = exif.ZoomStepCount;
   let fscinf = exif.FocusStepInfinity;
   let fsc = exif.FocusStepCount - (fscinf? fscinf: 0);
//console.log(exif.FocusStepCount,fscinf,fsc,zsc);

// 1
   if ( exif.DigitalZoomRatio == 0 ) {
    M.log('DigitalZoomRatio');
    return;
   }
//   M.log('ZoomStepCount',zsc);
//   M.log('FocusStepCount',fsc);

   let data = convertData(profile,['distance','zsc','fsc','fov','fscinf']);
   if ( !data ) {
    return;
   }
   data = data.sort((a,b)=>
    a.zsc != b.zsc? a.zsc - b.zsc: a.distance - b.distance
   );

//console.log(data);
   let zidx = {};
   for( let i = 0; i < data.length; i++ ) {
    let r = data[i];
    if ( fscinf ) { r.fsc -= r.fscinf; }
      // fscinf: Not stored when FourThirds Body + FourThirds Lens,
      //   althow stored when M.FourThirds Body + MMF-3 + FourThirds Lens.
// 未確認
    if ( zidx[r.zsc] === undefined ) { zidx[r.zsc] = []; }
    zidx[r.zsc].push(r);
   }
   let zlist = Object.keys(zidx).map(a=>parseFloat(a)).sort((a,b)=>a-b);
   if ( zlist.length < 2 ) return;
   let i = 0;
   for ( ; i < zlist.length-1 && zlist[i+1] <= zsc; i++ ) {}
   if ( i >= zlist.length-1 ) i--;
   let z0 = zlist[i], z1 = zlist[i+1];

// dlist...
   let dlist = [];
   let didx = {};
   for( let i = 0; i < zidx[z0].length; i++ ) {
    let r = zidx[z0][i];
    didx[r.distance] = r;
   }
   for( let i = 0; i < zidx[z1].length; i++ ) {
    let r = zidx[z1][i];
    if ( didx[r.distance] ) {
     dlist.push([didx[r.distance],r]);
    }
   }
   if ( dlist.length < 2 ) return;

   let j = 0;
   let dir = dlist[0][0].fsc < dlist[1][0].fsc? 1: -1;
   for( ; j < dlist.length-1; j++ ) {
    let p = dlist[j+1];
    let f = interpolate(p[0].zsc,p[0].fsc,p[1].zsc,p[1].fsc,zsc);
    if ( dir*f > dir*fsc ) { break; }
   }
   if ( j >= dlist.length-1 ) j--;
   let r00 = dlist[j][0], r01 = dlist[j][1];
   let r10 = dlist[j+1][0], r11 = dlist[j+1][1];
   let fsc0 = interpolate(r00.zsc,r00.fsc,r01.zsc,r01.fsc,zsc);
   let fsc1 = interpolate(r10.zsc,r10.fsc,r11.zsc,r11.fsc,zsc);
   let fov0, fov1;
   if ( profile.zoomCurve === 'log' ) {
      // log 
      fov0 = interpolateZLog(r00.zsc,r00.fov,r01.zsc,r01.fov,zsc);
      fov1 = interpolateZLog(r10.zsc,r10.fov,r11.zsc,r11.fov,zsc);
   } else {
      fov0 = interpolateZ(r00.zsc,r00.fov,r01.zsc,r01.fov,zsc);
      fov1 = interpolateZ(r10.zsc,r10.fov,r11.zsc,r11.fov,zsc);
   }
//console.log(fsc0,fsc1,fov0,fov1,fsc);
   let fov = interpolateF(fsc0,fov0,fsc1,fov1,fsc);
   return fov;
  }


  function compute(profile,exif){
   if ( !profile ) { return; }
   let fovBase;
   switch( profile.type ) {
   case 'fixed-magnification': fovBase = profile.data; break;
   case 'fixed-focal-length' : fovBase = fixedFocal(profile,exif); break;
   case 'zoom'               : fovBase = zoomLens(profile,exif); break;
   }
   if ( !fovBase ) return;

   // CropWidthとCropHeightはAspect比を変えても変わらないので
   // センサーサイズと考えよう...
   let sensorWidth = exif.CropWidth;
   let sensorHeight = exif.CropHeight;
   let imageWidth = exif.ImageWidth;
   let imageHeight = exif.ImageHeight;
// たぶん...
   let cropRatio = 1;
   if ( imageHeight/imageWidth > sensorHeight/sensorWidth ) {
    cropRatio = sensorHeight/sensorWidth*(imageWidth/imageHeight);
   }
   let fov_width_mm = fovBase*cropRatio/exif.DigitalZoomRatio;
   return { width: fov_width_mm };
  }




  function getKey(exif){
   let cameraCode = exif._BodyType||exif._CameraCode;
   if ( exif && exif._MakerCode && cameraCode ) {
    let key = [
     exif._MakerCode, cameraCode, exif._LensCode,
     exif._ExtenderCode, exif._ModeCode
    ].map(a=>a==null?'':encodeURI(a)).join(':');
    return key;
   }
  }




  function addProfile(profile){
   let key = [
    profile.maker,
    profile.bodyType||profile.cameraCode,
    profile.lensCode,
    profile.extenderCode,
    profile.mode
   ].map(a=>a==null?'':encodeURI(a)).join(':');
   if ( profiles[key] === undefined ) {
    profiles[key] = {};
   }
   let subkey = profile.attachment? profile.attachment: '';
   profiles[key][subkey] = profile;
  };
   

  function getProfiles(){
   return profiles;
  }


  function restoreProfiles(p){
/*
   if ( p ) {
    profiles = p;
   } else {
    profiles = {};
    BP.data.forEach(a=>addProfile(a));
   }
*/
    profiles = {};
    BP.data.forEach(a=>addProfile(a));
  }

  function reset(){
   profiles = {};
   BP.data.forEach(a=>addProfile(a));
  }

  function clear(){
   profiles = {};
  }

  function getList(exif){
   let key = getKey(exif);
   let set = profiles[key] || {};
   let list = Object.keys(set).sort().map(a=>set[a]);
   return list;
  }


  return Fov;
 });

})();
