(function(){
 'use strict';

 const Ld = sLoader;

 Ld.module('imageInfo',function(){

  const Lp   = Ld.use('layerParts');
  const M    = Ld.ref('main');
  const U    = Ld.use('utility');
  const Mp   = Ld.use('mainParam');
  const Mb   = Ld.use('messageBox');

  const Info = {};

  Info.show = function(){
   Mb.messageBox({
    title:'Image Info.',
    text: getText()
   });
  };

  function getText(){
   let exif = Mp.exif;
   let h = `<div class="help">
    <table class="padding" >
     <colgroup>
      <col style="width:40%;" />
      <col style="width:60%;" />
     </colgroup>
     <tbody>`;

   function tr(title,text){
    h += `<tr><td>${title}</td><td>${text}</td></tr>`;
   }

   function esc(s) { return U.htmlEscape(s); }

   let v, s;

   if ( exif.Make ) { tr('Make', esc(exif.Make)); }

   s = exif.CameraType || exif.Model;
   if ( exif.Model ) { tr('Camera Body', esc(s)); }
   if ( exif.LensType ) {
    tr('Lens', esc(exif.LensType));
   }
   if ( exif.Extender ) {
    tr('Extender' , esc(exif.Extender));
   }
   if ( exif.ConversionLens ) {
    tr('Conversion Lens' , esc(exif.ConversionLens));
   }

   if ( exif.DateTimeOriginal ) {
    tr('Date Time Original',esc(exif.DateTimeOriginal));
   }
// File Name

   // File size

   tr('Image Width',esc(exif.ImageWidth));
   tr('Image Height',esc(exif.ImageHeight));


   // setting
   if ( exif.FNumber ) {
    tr('F Number',esc(exif.FNumber));
   }
   v = exif.ExposureTime||exif.ShutterSpeedValue;
   if ( v !== undefined ) {
    s = v < 1? '1/'+U.roundRelative(1/v,3): U.roundRelative(v,3);
    tr('Shutter Speed', s);
   }
   if ( exif.ISO ) {
    tr('ISO',esc(exif.ISO));
   }

// FocalLengthIn35mmFormat
   if ( exif.FocalLength ) {
    tr('Focal Length',U.roundRelative(exif.FocalLength,3));
   }

   if ( !isNaN(exif.GPSLatitude) && !isNaN(exif.GPSLongitude) ) {
    let coord = exif.GPSLatitude.toFixed(6)
     + ', ' + exif.GPSLongitude.toFixed(6);
    tr('GPS Position',esc(coord));
   }
   if ( !isNaN(exif.GPSAltitude) ) {
    tr('GPS Altitude',esc(exif.GPSAltitude.toFixed(1)));
   }

   if ( exif._fovAlgorithm ) {
    tr('FOV Algorithm',esc(exif._fovAlgorithm));
   }
   if ( exif._fovParam ) {
    tr('FOV Parameter',esc(exif._fovParam));
   }
   if ( exif._Mode ) {
    tr('Mode',esc(exif._Mode));
   }
   if ( exif.DigitalZoomRatio ) {
    tr('Digital Zoom Ratio',esc(exif.DigitalZoomRatio));
   }
   if ( exif.Orientation ) {
    tr('Orientation',esc(exif.Orientation));
   }
   if ( exif.FocusStepCount ) {
    tr('Focus Step Count',esc(exif.FocusStepCount));
   }
   if ( exif.ZoomStepCount ) {
    tr('Zoom Step Count',esc(exif.ZoomStepCount));
   }
   if ( Mp.hfov ) {
    tr('FOV',esc(Mp.hfov));
   }



/*
   let list = Object.keys(exif).sort();
   for( let i = 0; i < list.length; i++ ) {
    let key = list[i];
    if ( !key.match(/^_/) ) {
     h += `<tr><td>${U.htmlEscape(key)}</td>
       <td>${U.htmlEscape(exif[key])}</td></tr>`;
    }
   }
*/
   h += `</tobdy></table></div>`;
   return h;
  }

  return Info;

 });

})();
