/*
 M.
  imageWidth
  hfov
  imageDiagonal
  cropWidth
  cropHeight

 U.
  measureText

*/
(function(){
 'use strict';
 const Ld = sLoader;
 Ld.module('scaleBar',function(){

  const Lp   = Ld.use('layerParts');
  const Sb = Lp.derive();
  Sb._className = 'ScaleBar';

  Sb.sig = 'ScaleBar';

  const M    = Ld.ref('main');
  const U    = Ld.use('utility');
  const Mn   = Ld.use('menu');
  const Icon = Ld.use('icon');
  const Mp   = Ld.use('mainParam');
  const Cp   = Ld.use('cropParam');
  const Shadow = Ld.use('filter-shadow');
  const Mb   = Ld.use('messageBox');
//  const Measure = Ld.use('measure');


  const baselineOffset = 0.2;  // ...

  const BAR_WIDTH_MIN = 0.1;  // %
  const BAR_WIDTH_MAX = 5.0;
  const BAR_FSIZE_MIN = 0.4;
  const BAR_FSIZE_MAX = 10.0;

  // 
  const PRESET_BAR_LENGTH = [
   "0.5mm",
   "1mm",
   "2mm"  ,
   "5mm",
   "10mm",
   "20mm",
   "50mm",
  ];

  Sb.defineMenu = function(){
   Sb._super.defineMenu.call(this);

//   let obj = this;

   this.sbmenu = {
    show          : Mn.visibilitySwitch({priority:90}),
    orientation   : Mn.barOrientation({priority:80}),
    transform       :
      Mn.CTransform.create({priority:75,
       buttonText: 'Move',
       x:{ min:-0.5, max:0.5, default:0.5},
       y:{ min:-0.5, max:0.5, default:0.5},
/*
       toInternalCoordinate(pos){
        let x = (pos.x - M.pRect.width/2) / M.cRect.width;
        let y = (pos.y - M.pRect.height/2) / M.cRect.height;
        if ( obj.values.orientation ) {
         return { x: y, y: -x };
        } else {
         return { x: x, y: y };
        }
       },
*/
       screenToInternalCoordinate:function(pos){
        let p = U.convertFromScreenCoordinate(M.paneEl,pos);
        let cw = M.scale*Cp.width;
        let ch = M.scale*Cp.height;
//
        let x = (p.x-cw/2)/cw;
        let y = (p.y-ch/2)/ch;
        if ( obj.values.orientation ) {
         return { x: y, y: -x };
        } else {
         return { x: x, y: y };
        }
       },
       ontransformstart:function(){ obj.ontransformstart(); },
       ontransformend:function(){ obj.ontransformend(); },
       onopen: function() { obj.ontransformopen(); },
       onclose: function() { obj.ontransformclose(); },
      }),
/*
    length        :
          Mn.CSelect.create({priority:72,
           buttonText:'Bar Length', title:'Bar Length',
           default:'5mm', list:PRESET_BAR_LENGTH,
          }),
*/
    length        :
          Mn.CText.create({priority:72,
           buttonText:'Bar Length',
           default:'5mm',
           historyKey:'scalebarLength',
          }),

/*
    hfov          :
          Mn.CText.create({priority:70,
           buttonText:'HFOV',
           historyKey: 'HFOV',
           placeholder:`Long side dimension of FOV. e.g. '23.6mm'`,
          }),
*/

    color         : Mn.bwColor({priority:69,default:0}),
    opacity       : Mn.opacity({priority:68,default:80}),
    backOpacity   : Mn.opacity({priority:67,buttonText:'backOpacity',
                  default:20}),
    lineWidth     : Mn.CSlider.create({priority:66, unit:'%',
                  curve:'log',
                  min:BAR_WIDTH_MIN, max:BAR_WIDTH_MAX, default:0.63,
                  buttonText:'Line Width', title:'Line Width'
                  }),

    fontSize      : Mn.CSlider.create({priority:65, unit:'%',
                  curve:'log',
                  min:BAR_FSIZE_MIN, max:BAR_FSIZE_MAX, default:2.5,
                  buttonText:'Font Size', title:'Font Size'
                  }),

// やっぱり自動でやるのはムリがあるなぁ。
    font          : Mn.font({priority:48}),
    fontWeight    : Mn.fontWeight({priority:47}),
    italic        : Mn.italicSwitch({priority:46}),
   };

//
   let obj = this;
   this.addMenu(this.sbmenu);
  };


  Sb.initialize = function(){
   Sb._super.initialize.call(this);
   this.values = this.menu.getValues();
// ?
   this.values.x = 0;
   this.values.y = 0;
  };


  Sb.onChange = function(values){
   if ( values.length !== undefined ) {
    let dim = parseDimension(values.length);
    if ( dim.value <= 0 ) {
     Mb.messageBox({
      title:'ERROR',
      text:`<p>Length must be a positive number.</p>`
     });
// リターンしても保存されてしまう...
// リストに保存されるので前でチェックしないと...
     return;
    }
    else if ( dim.unit == '' ) {
     Mb.messageBox({
      title:'ERROR',
      text:`<p>Available units are 'um'(micrometer), 'mm', 'cm', 'm', 'km'.</p>`
     });
     return;
    }
   }
   Sb._super.onChange.call(this,values);
  };

/*
  Sb.getSetting = function(){
   let v = Object.assign({},this.values);
   delete v.hfov;
   return v;
  };

  Sb.loadSetting = function(data){
   Sb._super.loadSetting.call(this,data);
   this.values.hfov = '';
   this.sbmenu.hfov.value = '';
console.log('loadSetting', data.hfov, this.values.hfov);
  };
*/

/*
  Sb.onOpenImage = function(){
   this.values.hfov = '';
   this.sbmenu.hfov.value = '';
  };
*/


  Sb.ontransformopen = function(){
   this.highlightStatus = true;
   this.dirty = 1;
   M.disp();
  };

  Sb.ontransformclose = function(){
   this.highlightStatus = false;
   this.dirty = 1;
   M.disp();
  };


  Sb.ontransformstart = function(){
//   console.log('ontransformstart');
   let rect = getRect(this.values);
   let cw, ch;
   if ( this.values.orientation ) {
    cw = Cp.height; ch = Cp.width;
   }
   else {
    cw = Cp.width; ch = Cp.height;
   }
   this.values.transform.x = rect.centerX/cw;
   this.values.transform.y = rect.centerY/ch;
   this.dirty = 1;
   M.disp();
  };

  Sb.ontransformend = function(){
//   console.log('ontransformend');
  };


  function parseDimension(str){
   let mm;
   let s = str.trim();
   let v = parseFloat(s);
   let unit = '';
// space
   let m = s.match(/([^\s\.0-9]+)$/);
   if ( m ) {
    unit = m[1];
    switch(unit){
    case 'μm':
    case 'µm' :
    case 'um' : mm = v/1000; unit = 'µm'; break; // U+00B5 MICRO SIGN
    case 'mm': mm = v; break;
    case 'cm': mm = v*10; break;
    case 'm' : mm = v*1000; break;
    case 'km': mm = v*1000000; break;
    default  : unit = ''; break;
    }
   }
//console.log('mm',mm);
   return {
    mm: mm,
    htmlText: v+'\u2006'+unit,
    value: v,
    unit: unit,
   };
  }


  Sb.getHighlightBox = function(){
   let box = getRect.call(this,this.values);
   if ( box && this.values.orientation ) {
    let t = box.x;
    box.x = -box.y-box.height; box.y = t;
    t = box.width; box.width = box.height; box.height = t;
   }
   return box;
  };


// getGeometry
  function getRect(V){

   let hfov = M.getHFOV();
/*
   if ( V.hfov !== undefined && V.hfov != '' ) {
    hfov = parseDimension(V.hfov).mm;
   }
   else {
    hfov = Mp.hfov;
   }
*/

   if ( !M.imageReady || !V.show || !hfov || hfov <= 0 ) {
//console.log('scalebar',M.imageReady,V.show,hfov);
    return;
   }
   const PADDING_TOP = 0.25;
   const PADDING_BOTTOM = 0.25;
   const MARGIN_TB = 0.1;
   const PADDING_LEFT = 0.5;
   const PADDING_RIGHT = 0.5;

   let sizeBase = 1;
   let lenMm = parseDimension(V.length).mm;

// barLength
   let len = Mp.imageWidth * lenMm/hfov
     * Cp.transform.zoom/Mp.imageDiagonal;

//
   let fontSize = 100; // for font size limitation
   let metrics = U.measureText(
    parseDimension(V.length).htmlText,
    { fontFamily:V.font, fontWeight:V.fontWeight*100,
      fontStyle: V.italic? 'italic': 'normal',
      fontSize: fontSize+'px' }
   );

// fontSize (internal unit)
   let fs = sizeBase*V.fontSize/100;
   let lw = sizeBase*V.lineWidth/100;

//   let fsScale = fontSize/fs;
   let fsScale = fs/fontSize;

// textExtent
   let textlen = metrics.width*fsScale;
   if (textlen > len) {
    fsScale *= len/textlen;
    fs *= len/textlen;
    lw *= len/textlen;
    textlen = len;
   }
   let textHeight = metrics.height*fsScale;

 // 最大の lw/len

   let bw = len + (textHeight+lw)*(PADDING_LEFT+PADDING_RIGHT);
   let bh = lw+textHeight+(textHeight+lw)*(PADDING_TOP+PADDING_BOTTOM+MARGIN_TB);
   let cw, ch;
   if ( V.orientation ) {
    cw = Cp.height/Cp.diagonal;
    ch = Cp.width/Cp.diagonal;
   } else {
    cw = Cp.width/Cp.diagonal;
    ch = Cp.height/Cp.diagonal;
   }
   let x = V.transform.x*cw - bw/2;
   let y = V.transform.y*ch - bh/2;
   if ( x < -cw/2 ) { x = -cw/2; }
   if ( x + bw > cw/2 ) { x = cw/2 - bw; }
   if ( y < -ch/2 ) { y = -ch/2; }
   if ( y + bh > ch/2 ) { y = ch/2 - bh; }
   return {
    x           : x,
    y           : y,
    width       : bw,
    height      : bh,  // box
    textX       : x+(textHeight+lw)*PADDING_LEFT+(len-textlen)/2,
    textY       : y+(textHeight+lw)*PADDING_TOP+
                   textHeight - metrics.baselineOffset*fsScale,
    barX        : x+(textHeight+lw)*PADDING_LEFT,
    barY        : y+bh-(textHeight+lw)*PADDING_BOTTOM-lw/2,
    barLen      : len,
    centerX     : x+bw/2,
    centerY     : y+bh/2,
    fontSize    : fs,
    lineWidth   : lw,
   };
  }


  // scale
  // 小さいとだめだ。Safariで。スケールに対応してないパラメタがあるみたい。
  // -> resolutionとかの設定をしてなかったからかも。
  // まぁ、pixel単位で処理することに
  //
  //  とりあえず、feGaussianBlurの stdDeviation
  // 適当な大きさ 1000 ~ 2000ぐらいにしておくのが無難かも
  // 掛けないとだめか。
  // もう一度確認
  // やっぱりだめだ。
  // 1000ぐらいにしておくか。
  // width, heightを入れないとFirefoxでおかしくなる。...
  // 

  Sb.makeSvgText = function(final,p){
   let V = this.values;
   let svg = "";
   let box = getRect(V);
   if ( box ) {

    let scale = p.cRect.diagonal;
//scale = 1000;
    let x = scale*(box.x-box.width*0.2);
    let y = scale*(box.y-box.height*0.2);
    let width = scale*(box.width*1.4);
    let height = scale*(box.height*1.4);
    if ( V.orientation ) {
     let t = x;
     x = -y-height; y = t;
     t = width; width = height; height = t;
    }
    let cw = scale*Cp.width;
    let ch = scale*Cp.height;
    svg = `<svg viewBox="${-cw/2} ${-ch/2} ${cw} ${ch}"
      x="0" y="0" width="100%" height="100%"
      style="position:absolute; left:0; top:0;"
      opacity="${V.opacity/100}"
      xmlns="http://www.w3.org/2000/svg"
      xmlns:xlink="http://www.w3.org/1999/xlink"
     >
    `;
    svg += `
     <defs>
      <filter id="blur"
      >
        <feGaussianBlur stdDeviation="${4/1000*scale}" />
      </filter>
     </defs>
    `;

    svg += `<g ${V.orientation? 'transform="rotate(90)"': ''}>`;

    // back rect
    svg += `
     <g >
      <rect x="${box.x*scale}" y="${box.y*scale}"
       width="${box.width*scale}" height="${box.height*scale}"
       fill="${V.color? "black": "white"}"
       fill-opacity="${V.backOpacity/100}"
      />
     </g>
    `;

    svg += `<g >`;
    
    // bar
    svg += `
     <path d="M ${scale*box.barX} ${scale*box.barY} 
      L ${scale*(box.barX+box.barLen)} ${scale*box.barY} z"
      stroke="${V.color? "white": "black"}"
      stroke-width="${scale*box.lineWidth}"
      stroke-opacity="1"
     />
    `;

    // text
    svg += `
     <text
      x="${scale*box.textX}"
      y="${scale*box.textY}"
      font-family="${V.font}"
      font-size="${scale*box.fontSize}"
      ${V.italic? ' font-style="italic"': ''}
      font-weight="${V.fontWeight*100}"
      fill="${V.color? "white": "black"}"
      fill-opacity="1"
     >
      ${parseDimension(V.length).htmlText}
     </text>
    `;
    // barLength escapeしといたほうがいいか..

    svg += '</g>';
    svg += `</g></svg>`;
   }
   return svg;
  };

  return Sb;

 });
})();
