/*
 *  wm-text.js
 */
(function(){
 'use strict';
 const Ld = sLoader;
 Ld.module('wm-text',function(){
  const Wm = Ld.use('waterMark');
  const WmT = Wm.derive();

  const M  = Ld.ref('main');
  const U  = Ld.use('utility');
  const Mn = Ld.use('menu');
  const Cp = Ld.use('cropParam');

  WmT._className = "wmText";
  WmT.sig = 'WMText';

  WmT.defineMenu = function(){
   WmT._super.defineMenu.call(this);

   this.addMenu({
    text          : Mn.CText.create({ priority:20,
                     buttonText:'Text',
                     default:'John Doe',
                     historyKey:'watermarkText',
                    }),
    font          : Mn.font({priority:10}),
    fontWeight    : Mn.fontWeight({priority:5}),
    italic        : Mn.italicSwitch({priority:3}),
   });

  };

  WmT.getSourceBoundingBox = function(){
   let geom = this.getGeometry();
   if ( !geom ) { return; }
   const MARGIN = 0.2;
   return{
    x     : geom.x - geom.width*MARGIN,
    y     : geom.y - geom.height*MARGIN,
    width : geom.width*(1+2*MARGIN),
    height: geom.height*(1+2*MARGIN),
   };
  };

  WmT.getHighlightBox = function(){
   return this.getGeometry();
  };

  WmT.getGeometry = function(){
   let V = this.values;
   if ( !M.imageReady || !V.show || V.text === undefined ) { return undefined; }
   let sizeBase = Cp.width/Cp.diagonal;
   let size = sizeBase*V.transform.zoom/100;
   let fontSize = 100; // px 仮の値
   let metrics = U.measureText(
    U.htmlEscape(V.text),
    { fontFamily: V.font, fontWeight: V.fontWeight*100,
      fontStyle: V.italic? 'italic': 'normal',
     fontSize: fontSize+'px',
     }
   );
   if ( metrics.width <= 0 ) {
    return undefined;
   }
   let scale = size/(metrics.width+fontSize*0.2);
   let fs = fontSize*scale;
   let bw = V.transform.zoom*sizeBase/100;
   let bh = metrics.height*scale;
   let x = V.transform.x /* *Cp.width/Cp.diagonal */ - bw/2;
   let y = V.transform.y /* *Cp.height/Cp.diagonal */ - bh/2;
   let cw = Cp.width/Cp.diagonal/2;
   let ch = Cp.height/Cp.diagonal/2;
   if ( x < -cw ) { x = -cw; }
   if ( x + bw > cw ) { x = cw - bw; }
   if ( y < -ch ) { y = -ch; }
   if ( y + bh > ch ) { y = ch - bh; }
   return {
    x           : x,
    y           : y,
    width       : bw,
    height      : bh,
    textX       : x+fontSize*0.1*scale,
    textY       : y+bh-metrics.baselineOffset*scale,
    fontSize    : fs,
    centerX     : x+bw/2,
    centerY     : y+bh/2
   };
  };


  WmT.makeSvgTextContent = function(final,scale){
   let geom = this.getGeometry();
   let V = this.values;
   let svg = `
    <text
     x="${scale*geom.textX}"
     y="${scale*geom.textY}"
     font-family="${V.font}"
     font-size="${scale*geom.fontSize}"
     ${V.italic? ' font-style="italic"': ''}
     font-weight="${V.fontWeight*100}"
     fill="white"
    >
     ${U.htmlEscape(V.text)}
    </text>
   `;
   return svg;
  };


  WmT.onChange = function(values){
   if ( values.text !== undefined ) {
    this.menuDefs.font.setProperties({
     sampleText: values.text
    });
   }
   WmT._super.onChange.call(this,values);
  };

  WmT.restore = function(data){
   WmT._super.restore.call(this,data);
   if ( this.values.text !== undefined ) {
    this.menuDefs.font.setProperties({
     sampleText: this.values.text
    });
   }
  };
// restoreSessionStatus
// loadSetting

  return WmT;
 });
})();
